!> @file user_actions.f90
!--------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the terms
! of the GNU General Public License as published by the Free Software Foundation,
! either version 3 of the License, or (at your option) any later version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2014 Leibniz Universitaet Hannover
!--------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: user_actions.f90 1683 2015-10-07 23:57:51Z knoop $
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1353 2014-04-08 15:21:23Z heinze
! REAL constants provided with KIND-attribute
! 
! 1320 2014-03-20 08:40:49Z raasch
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements 
!
! 1318 2014-03-17 13:35:16Z raasch
! module interfaces removed
!
! 1053 2012-11-13 17:11:03Z hoffmann
! +qr-tendency, nr-tendency
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 211 2008-11-11 04:46:24Z raasch
! Former file user_interface.f90 split into one file per subroutine
!
! Description:
! ------------
!> Execution of user-defined actions before or after single timesteps
!------------------------------------------------------------------------------!
 MODULE user_actions_mod
 

    PRIVATE
    PUBLIC user_actions

    INTERFACE user_actions
       MODULE PROCEDURE user_actions
       MODULE PROCEDURE user_actions_ij
    END INTERFACE user_actions

 CONTAINS


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE user_actions( location )

       USE control_parameters

       USE constants,                                                          &
           ONLY:  pi

       USE cpulog

       USE grid_variables

       USE indices

       USE kinds

       USE pegrid

       USE user

       USE arrays_3d,                                                          &
           ONLY:  ddzu, dzu, dzw, tend, u, v, w

       USE MPI

       IMPLICIT NONE

       CHARACTER (LEN=*) ::  location !< 

       INTEGER(iwp) ::  i !< 
       INTEGER(iwp) ::  j !< 
       INTEGER(iwp) ::  k !< 
       INTEGER(iwp) ::  m !< 
       INTEGER(iwp) ::  n !< 

       REAL(wp) :: ff_ref            !< reference wind speed
       REAL(wp) :: o                 !< for checking wind directions
       REAL(wp) :: u_centre          !< u at cell centre
       REAL(wp) :: v_centre          !< v at cell centre
       REAL(wp) :: vol_flow_l        !< local volume flow
       REAL(wp) :: w_centre          !< w at cell centre

       CALL cpu_log( log_point(24), 'user_actions', 'start' )

!
!--    Here the user-defined actions follow
!--    No calls for single grid points are allowed at locations before and
!--    after the timestep, since these calls are not within an i,j-loop
       SELECT CASE ( location )

          CASE ( 'before_timestep' )
!
!--          Enter actions to be done before every timestep here


          CASE ( 'after_integration' )



!-- Calculation of horizontal wind and horizontal wind direction on zw-grid!!!
!-- keep in mind that this calculation is done for ALL grid points -- some other
!-- heights may be needed in user_statistics for virtual measuring stations -- 
!-- which is NOT very cpu-optimized...
          DO  i = nxl, nxr
             DO  j = nys, nyn
                DO  k = nzb_s_inner(j,i)+1, nzt
                   ui(k,j,i) = 0.25_wp * ( u(k,j,i) + u(k,j,i+1) + u(k+1,j,i) + u(k+1,j,i+1))
                   vi(k,j,i) = 0.25_wp * ( v(k,j,i) + v(k,j+1,i) + v(k+1,j,i) + v(k+1,j+1,i))
                ENDDO
             ENDDO
          ENDDO

          CALL exchange_horiz( ui, nbgp )
          CALL exchange_horiz( vi, nbgp )

          DO  i = nxlg, nxrg
             DO  j = nysg, nyng
                DO  k = nzb_s_inner(j,i)+1, nzt
                   u2(k,j,i) = ui(k,j,i)**2
                   v2(k,j,i) = vi(k,j,i)**2
                   uvgs(k,j,i) = SQRT( u2(k,j,i) + v2(k,j,i) )
                   dd(k,j,i)   = ATAN2( -ui(k,j,i), -vi(k,j,i) ) &
                                 * 180.0_wp / pi + 180.0_wp
                ENDDO
             ENDDO
          ENDDO


!
!--          Calculation of output quantities
             DO  i = nxl, nxr
                DO  j = nys, nyn

!--                Reference wind speed for velocity ratio
                   k = nzt
                   u_centre = ( u(k,j,i) + u(k,j,i+1) ) * 0.5
                   v_centre = ( v(k,j,i) + v(k,j+1,i) ) * 0.5
                   w_centre = ( w(k,j,i) + w(k-1,j,i) ) * 0.5

                   ff_ref = SQRT( u_centre**2 + v_centre**2 + w_centre**2 )

                   DO  k = nzb_s_inner(j,i)+1, nzt

!--                   Wind speed, direction, velocity ratio at scalar grid point
                      u_centre = ( u(k,j,i) + u(k,j,i+1) ) * 0.5
                      v_centre = ( v(k,j,i) + v(k,j+1,i) ) * 0.5
                      w_centre = ( w(k,j,i) + w(k-1,j,i) ) * 0.5

!                       dd(k,j,i) = ATAN2( u_centre, v_centre )
                      ff(k,j,i) = SQRT( u_centre**2 + v_centre**2 + w_centre**2 )
                      ff_r(k,j,i) = ff(k,j,i) / ff_ref

!--                   Calculate turbulence intensity from rot[(u,v,w)]
!--                   at scalar grid point
                      ti(k,j,i) = 0.25_wp * SQRT(                           &
                        (   (   w(k,j+1,i) + w(k-1,j+1,i)                   &
                              - w(k,j-1,i) - w(k-1,j-1,i) ) * ddy           &
                          - (   v(k+1,j,i) + v(k+1,j+1,i)                   &
                              - v(k-1,j,i) - v(k-1,j+1,i) ) * ddzu(k) )**2  &
                      + (   (   u(k+1,j,i) + u(k+1,j,i+1)                   &
                              - u(k-1,j,i) - u(k-1,j,i+1) ) * ddzu(k)       &
                          - (   w(k,j,i+1) + w(k-1,j,i+1)                   &
                              - w(k,j,i-1) - w(k-1,j,i-1) ) * ddx     )**2  &
                      + (   (   v(k,j,i+1) + v(k,j+1,i+1)                   &
                              - v(k,j,i-1) - v(k,j+1,i-1) ) * ddx           &
                          - (   u(k,j+1,i) + u(k,j+1,i+1)                   &
                              - u(k,j-1,i) - u(k,j-1,i+1) ) * ddy     )**2  )
                   ENDDO

                ENDDO
             ENDDO

             CALL exchange_horiz( ff, nbgp )
             CALL exchange_horiz( ff_r, nbgp )
             CALL exchange_horiz( ti, nbgp )


          CASE ( 'after_timestep' )


          CASE ( 'u-tendency' )
!
!--          Enter actions to be done in the u-tendency term here


          CASE ( 'v-tendency' )


          CASE ( 'w-tendency' )


          CASE ( 'pt-tendency' )


          CASE ( 'sa-tendency' )


          CASE ( 'e-tendency' )


          CASE ( 'q-tendency' )

             IF ( passive_scalar ) THEN
                DO  i = nxl, nxr
                   DO  j = nys, nyn
                      tend(nzb_s_inner(j,i)+1,j,i) =                           &
                         tend(nzb_s_inner(j,i)+1,j,i) + s_emission_rate(j,i)   &
                         * dx * dy * dzu(nzb_s_inner(j,i)+1)
                   ENDDO
                ENDDO
             ENDIF

          CASE ( 'qr-tendency' )


          CASE ( 'nr-tendency' )


          CASE DEFAULT
             message_string = 'unknown location "' // location // '"'
             CALL message( 'user_actions', 'UI0001', 1, 2, 0, 6, 0 )

       END SELECT

       CALL cpu_log( log_point(24), 'user_actions', 'stop' )

    END SUBROUTINE user_actions


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for grid point i,j
!------------------------------------------------------------------------------!
    SUBROUTINE user_actions_ij( i, j, location )

       USE arrays_3d,                                                          &
           ONLY:  dzu, tend

       USE control_parameters

       USE grid_variables,                                                     &
           ONLY:  dx, dy

       USE indices,                                                            &
           ONLY:  nzb, nzb_s_inner

       USE kinds
       USE pegrid
       USE user

       IMPLICIT NONE

       CHARACTER (LEN=*) ::  location

       INTEGER(iwp) ::  i
       INTEGER(iwp) ::  idum
       INTEGER(iwp) ::  j
       INTEGER(iwp) ::  k

!
!--    Here the user-defined actions follow
       SELECT CASE ( location )

          CASE ( 'u-tendency' )
!
!--          Enter actions to be done in the u-tendency term here


          CASE ( 'v-tendency' )


          CASE ( 'w-tendency' )


          CASE ( 'pt-tendency' )


          CASE ( 'sa-tendency' )


          CASE ( 'e-tendency' )


          CASE ( 'q-tendency' )

             IF ( passive_scalar ) THEN
                tend(nzb_s_inner(j,i)+1,j,i) = tend(nzb_s_inner(j,i)+1,j,i)    &
                   + s_emission_rate(j,i) * dx * dy * dzu(nzb_s_inner(j,i)+1)
             ENDIF

          CASE ( 'qr-tendency' )


          CASE ( 'nr-tendency' )


          CASE ( 'before_timestep', 'after_integration', 'after_timestep' )
             message_string = 'location "' // location // '" is not ' // &
                             'allowed to be called with parameters "i" and "j"'
             CALL message( 'user_actions', 'UI0002', 1, 2, 0, 6, 0 )


          CASE DEFAULT
             message_string = 'unknown location "' // location // '"'
             CALL message( 'user_actions', 'UI0001', 1, 2, 0, 6, 0 )
             

       END SELECT

    END SUBROUTINE user_actions_ij

 END MODULE user_actions_mod
