!> @file user_additional_routines.f90
!--------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the terms
! of the GNU General Public License as published by the Free Software Foundation,
! either version 3 of the License, or (at your option) any later version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2014 Leibniz Universitaet Hannover
!--------------------------------------------------------------------------------!
!
! Current revisions:
! -----------------
! 
! 
! Former revisions:
! -----------------
! $Id: user_additional_routines.f90 1683 2015-10-07 23:57:51Z knoop $
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1320 2014-03-20 08:40:49Z raasch
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements 
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 211 2008-11-11 04:46:24Z raasch
! Initial version
!
! Description:
! ------------
!> This file may contain additional user-defined routines.
!------------------------------------------------------------------------------!
 SUBROUTINE user_dummy
 

    USE kinds

    USE user

    IMPLICIT NONE

!
!-- Here comes the code of the user-defined routine

 END SUBROUTINE user_dummy



 SUBROUTINE outflow_turbulence

!------------------------------------------------------------------------------!
! Current revisions:
! -----------------
! 2015-11-19 gronemeier
! Code annotations made doxygen readable.
!
! 2015-10-19 gronemeier
! either calcluate averaged profiles at source plane and use these as outflow
! values or use instantaneous values from source plane.
!
! 2015-09-22 gronemeier
! initial version
!
! Description:
! ------------
!> Routine based on inflow_turbulence.f90 (r1616). Copies values or averaged
!> profiles of 3d data from a source plane (defined by outflow_source_plane) to
!> the outflow boundary.
!------------------------------------------------------------------------------!

    USE arrays_3d,                                                             &
        ONLY:  e, pt, q, u, v, w

    USE control_parameters,                                                    &
        ONLY:  humidity, passive_scalar

    USE cpulog,                                                                &
        ONLY:  cpu_log, log_point

    USE grid_variables,                                                        &
        ONLY:  dx

    USE indices,                                                               &
        ONLY:  nbgp, nx, nxr, ny, nyn, nys, nyng, nysg, nzb, nzt

    USE kinds

    USE pegrid

    USE user,                                                                  &
        ONLY:  id_outflow, id_source_out, outflow_source_plane,                &
               turbulent_outflow_mean

    IMPLICIT NONE

    INTEGER(iwp) ::  i        !<
    INTEGER(iwp) ::  j        !<
    INTEGER(iwp) ::  k        !<
    INTEGER(iwp) ::  l        !<
    INTEGER(iwp) ::  next     !<
    INTEGER(iwp) ::  ngp_ifd  !<
    INTEGER(iwp) ::  ngp_pr   !<
    INTEGER(iwp) ::  prev     !<

    REAL(wp), DIMENSION(nzb:nzt+1,6,nbgp)           ::                         &
       avpr, avpr_l           !< averaged profiles (global, and local)

    REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,6,nbgp) ::                         &
       outflow_val            !< values to be copied to the outflow boundary


    CALL cpu_log( log_point(50), 'outflow_turbulence', 'start' )
!
!-- Carry out spanwise averaging in the source plane
    ngp_pr = ( nzt - nzb + 2 ) * 6 * nbgp
    ngp_ifd = ngp_pr * ( nyn - nys + 1 + 2 * nbgp )
!
!-- Get the values at the source plane
    i = outflow_source_plane / dx

!
!-- Carry out spanwise averaging in the source plane
    IF ( turbulent_outflow_mean )  THEN

       avpr_l = 0.0_wp

#if defined( __parallel )
       IF ( myidx == id_source_out )  THEN
!
!--       Summarise values in the source plane
          DO  l = 1, nbgp
             DO  j = nys, nyn
                DO  k = nzb, nzt + 1

                   avpr_l(k,1,l) = avpr_l(k,1,l) + u(k,j,i)
                   avpr_l(k,2,l) = avpr_l(k,2,l) + v(k,j,i)
                   avpr_l(k,3,l) = avpr_l(k,3,l) + w(k,j,i)
                   avpr_l(k,4,l) = avpr_l(k,4,l) + pt(k,j,i)
                   avpr_l(k,5,l) = avpr_l(k,5,l) + e(k,j,i)
                   IF ( humidity  .OR.  passive_scalar )                       &
                      avpr_l(k,6,l) = avpr_l(k,6,l) + q(k,j,i)

                ENDDO
             ENDDO
             i = i + 1
          ENDDO

       ENDIF
!
!--    Now, averaging over all PEs
       IF ( collective_wait )  CALL MPI_BARRIER( comm2d, ierr )
       CALL MPI_ALLREDUCE( avpr_l(nzb,1,1), avpr(nzb,1,1), ngp_pr, MPI_REAL,   &
                           MPI_SUM, comm2d, ierr )

#else
       DO  l = 1, nbgp
          DO  j = nys, nyn
             DO  k = nzb, nzt + 1

                avpr_l(k,1,l) = avpr_l(k,1,l) + u(k,j,i)
                avpr_l(k,2,l) = avpr_l(k,2,l) + v(k,j,i)
                avpr_l(k,3,l) = avpr_l(k,3,l) + w(k,j,i)
                avpr_l(k,4,l) = avpr_l(k,4,l) + pt(k,j,i)
                avpr_l(k,5,l) = avpr_l(k,5,l) + e(k,j,i)
                IF ( humidity  .OR.  passive_scalar )                          &
                   avpr_l(k,6,l) = avpr_l(k,6,l) + q(k,j,i)

             ENDDO
          ENDDO
          i = i + 1
       ENDDO

       avpr = avpr_l
#endif

       avpr = avpr / ( ny + 1 )
!
!--    Save averaged profiles as outflow values
       DO  l = 1, nbgp
          DO  j = nysg, nyng
             outflow_val(:,j,1,l) = avpr(:,1,l)
             outflow_val(:,j,2,l) = avpr(:,2,l)
             outflow_val(:,j,3,l) = avpr(:,3,l)
             outflow_val(:,j,4,l) = avpr(:,4,l)
             outflow_val(:,j,5,l) = avpr(:,5,l)
             outflow_val(:,j,6,l) = avpr(:,6,l)
          ENDDO
       ENDDO

!
!-- Use instantaneous values instead of averaged profiles
    ELSE

#if defined( __parallel )
       IF ( myidx == id_source_out )  THEN
          DO  l = 1, nbgp
             DO  j = nysg, nyng
                DO  k = nzb, nzt + 1

                   outflow_val(k,j,1,l) = u(k,j,i)
                   outflow_val(k,j,2,l) = v(k,j,i)
                   outflow_val(k,j,3,l) = w(k,j,i)
                   outflow_val(k,j,4,l) = pt(k,j,i)
                   outflow_val(k,j,5,l) = e(k,j,i)
                   IF ( humidity  .OR.  passive_scalar )                       &
                      outflow_val(k,j,6,l) = q(k,j,i)
               ENDDO
             ENDDO
             i = i + 1
          ENDDO

       ENDIF
#else
       DO  l = 1, nbgp
          DO  j = nysg, nyng
             DO  k = nzb, nzt+1

                outflow_val(k,j,1,l) = u(k,j,i)
                outflow_val(k,j,2,l) = v(k,j,i)
                outflow_val(k,j,3,l) = w(k,j,i)
                outflow_val(k,j,4,l) = pt(k,j,i)
                outflow_val(k,j,5,l) = e(k,j,i)
                IF ( humidity  .OR.  passive_scalar )                          &
                   outflow_val(k,j,6,l) = q(k,j,i)
             ENDDO
          ENDDO
          i = i + 1
       ENDDO
#endif

    ENDIF

!
!-- For parallel runs, send the values to the respective outflow PE
#if defined( __parallel )
    IF ( myidx == id_source_out  .AND.  myidx /= id_outflow )  THEN

       CALL MPI_SEND( outflow_val(nzb,nysg,1,1), ngp_ifd, MPI_REAL,            &
                      id_outflow, 1, comm1dx, ierr )

    ELSEIF ( myidx /= id_source_out  .AND.  myidx == id_outflow )  THEN

       outflow_val = 0.0_wp
       CALL MPI_RECV( outflow_val(nzb,nysg,1,1), ngp_ifd, MPI_REAL,            &
                      id_source_out, 1, comm1dx, status, ierr )

    ENDIF
#endif

!
!-- Copy values at the outflow
    IF ( nxr == nx )  THEN

       DO  j = nysg, nyng
          DO  k = nzb, nzt + 1

             u(k,j,nx+1:nx+nbgp)  = outflow_val(k,j,1,1:nbgp)
             v(k,j,nx+1:nx+nbgp)  = outflow_val(k,j,2,1:nbgp)
             w(k,j,nx+1:nx+nbgp)  = outflow_val(k,j,3,1:nbgp)
             pt(k,j,nx+1:nx+nbgp) = outflow_val(k,j,4,1:nbgp)
             e(k,j,nx+1:nx+nbgp)  = outflow_val(k,j,5,1:nbgp)
             e(k,j,nx+1:nx+nbgp)  = MAX( e(k,j,nx+1:nx+nbgp), 0.0_wp )

             IF ( humidity  .OR.  passive_scalar )                             &
                q(k,j,nx+1:nx+nbgp)  = outflow_val(k,j,6,1:nbgp)

          ENDDO
       ENDDO


    ENDIF


    CALL cpu_log( log_point(50), 'outflow_turbulence', 'stop' )


 END SUBROUTINE outflow_turbulence
